/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Abs_Poly_Adapter_hh
#define pplite_Abs_Poly_Adapter_hh 1

#include "Abs_Poly.hh"

#include "Poly.hh"
#include "F_Poly.hh"
#include "U_Poly.hh"
#include "Poly_Stats.hh"
#include "memory_in_bytes.hh"

#include <cassert>
#include <iostream>
#include <string>
#include <utility>

namespace pplite {
namespace dynamic {

// A template for a polymorphic sequence
template <typename Value, typename Sequence>
struct Abs_Sequence_Adapter : public Abs_Sequence<Value> {
private:
  using Base = Abs_Sequence<Value>;
  using seq_type = Sequence;
  seq_type seq;

public:
  using value_type = typename Base::value_type;

  explicit Abs_Sequence_Adapter(seq_type&& s) : seq(std::move(s)) {}
  ~Abs_Sequence_Adapter() override {}

  Abs_Sequence_Adapter(const Abs_Sequence_Adapter&) = delete;
  Abs_Sequence_Adapter(Abs_Sequence_Adapter&&) = delete;
  Abs_Sequence_Adapter& operator=(const Abs_Sequence_Adapter&) = delete;
  Abs_Sequence_Adapter& operator=(Abs_Sequence_Adapter&&) = delete;

  // Note: size() may be less than end_pos() due to skippable values.
  dim_type end_pos() const override { return seq.end_pos(); }
  dim_type size() const override { return seq.size(); }
  bool is_skippable(dim_type pos) const override {
    return seq.is_skippable(pos);
  }
  const value_type* get_value_ptr(dim_type pos) const override {
    return seq.get_value_ptr(pos);
  }
}; // Abs_Sequence_Adapter

template <typename PH>
class Abs_Poly_Adapter : public Abs_Poly {
public:
  using poly_type = PH;
  poly_type ph;

  // Helper
  static const poly_type& get_poly(const Abs_Poly& y) {
    assert(dynamic_cast<const Abs_Poly_Adapter*>(&y));
    return static_cast<const Abs_Poly_Adapter&>(y).ph;
  }

  explicit Abs_Poly_Adapter(const poly_type& y_ph) : ph(y_ph) {}
  explicit Abs_Poly_Adapter(poly_type&& y_ph) : ph(std::move(y_ph)) {}

  ~Abs_Poly_Adapter() = default;
  Abs_Poly_Adapter(const Abs_Poly_Adapter&) = delete;
  Abs_Poly_Adapter(Abs_Poly_Adapter&&) = delete;
  Abs_Poly_Adapter& operator=(const Abs_Poly_Adapter&) = delete;
  Abs_Poly_Adapter& operator=(Abs_Poly_Adapter&&) = delete;

  // Public interface
  explicit Abs_Poly_Adapter(dim_type d, Spec_Elem s, Topol t)
    : ph(d, s, t) {}

  Abs_Poly_Adapter* clone() const override {
    return new Abs_Poly_Adapter(ph);
  }

  void minimize() const override { ph.minimize(); }
  bool is_necessarily_closed() const override {
    return ph.is_necessarily_closed();
  }
  bool check_inv() const override { return ph.check_inv(); };
  bool is_empty() const override { return ph.is_empty(); }
  bool is_universe() const override { return ph.is_universe(); }
  bool is_bounded() const override { return ph.is_bounded(); }
  bool is_bounded_expr(bool from_below, const Linear_Expr& expr) const override {
    return ph.is_bounded_expr(from_below, expr);
  }
  bool is_topologically_closed() const override {
    return ph.is_topologically_closed();
  }
  bool boxed_contains(const Abs_Poly& y) const override {
    return ph.boxed_contains(get_poly(y));
  }
  bool constrains(Var v) const override { return ph.constrains(v); }
  bool contains(const Abs_Poly& y) const override {
    return ph.contains(get_poly(y));
  }
  bool equals(const Abs_Poly& y) const override {
    return ph.equals(get_poly(y));
  }
  bool is_disjoint_from(const Abs_Poly& y) const override {
    return ph.is_disjoint_from(get_poly(y));
  }

  Topol topology() const override { return ph.topology(); }
  dim_type space_dim() const override { return ph.space_dim(); }
  dim_type affine_dim() const override { return ph.affine_dim(); }
  dim_type num_min_cons() const override { return ph.num_min_cons(); }
  dim_type num_min_gens() const override { return ph.num_min_gens(); }
  Cons copy_cons() const override { return ph.copy_cons(); }
  Gens copy_gens() const override { return ph.copy_gens(); }
  Poly_Con_Rel relation_with(const Con& c) const override {
    return ph.relation_with(c);
  }
  Poly_Gen_Rel relation_with(const Gen& g) const override {
    return ph.relation_with(g);
  }
  BBox get_bounding_box() const override { return ph.get_bounding_box(); }
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& inf_n, Integer& inf_d,
           bool& minimum, Gen& g) const override {
    return ph.min(expr, inhomo, inf_n, inf_d, minimum, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& sup_n, Integer& sup_d,
           bool& maximum, Gen& g) const override {
    return ph.max(expr, inhomo, sup_n, sup_d, maximum, g);
  }

  bool
  get_bounds(Var var, Rational& min, Rational& max,
             bool& min_inf, bool& max_inf) const override{
    return ph.get_bounds(var, min, max, min_inf, max_inf);
  }

  size_t hash() const override { return ph.hash(); }
  size_t get_memory_in_bytes() const override {
    return pplite::total_memory_in_bytes(ph);
  }

  // Proxies.
  Cons_Proxy cons() const override {
    // Build proxy from a concrete sequence adapter.
    using Seq = Abs_Sequence_Adapter<Con, typename poly_type::Cons_Proxy>;
    return Cons_Proxy(new Seq(ph.cons()));
  }
  Cons_Proxy normalized_cons() const override {
    // Build proxy from a concrete sequence adapter.
    using Seq = Abs_Sequence_Adapter<Con, typename poly_type::Cons_Proxy>;
    return Cons_Proxy(new Seq(ph.normalized_cons()));
  }
  Gens_Proxy gens() const override {
    // Build proxy from a concrete sequence adapter.
    using Seq = Abs_Sequence_Adapter<Gen, typename poly_type::Gens_Proxy>;
    return Gens_Proxy(new Seq(ph.gens()));
  }

  // Modifiers
  void m_swap(Abs_Poly& y) noexcept override {
    assert(dynamic_cast<Abs_Poly_Adapter*>(&y));
    ph.m_swap(static_cast<Abs_Poly_Adapter&>(y).ph);
  }
  void set_empty() override { ph.set_empty(); }
  void set_universe() override { ph.set_universe(); }
  void set_topology(Topol t) override { ph.set_topology(t); }
  void add_con(const Con& c) override { ph.add_con(c); }
  void add_cons(const Cons& cs) override { ph.add_cons(cs); }
  void add_gen(const Gen& g) override { ph.add_gen(g); }
  void add_gens(const Gens& gs) override { ph.add_gens(gs); }

  void con_hull_assign(const Abs_Poly& y, bool boxed) override {
    ph.con_hull_assign(get_poly(y), boxed);
  }
  void concatenate_assign(const Abs_Poly& y) override {
    ph.concatenate_assign(get_poly(y));
  }
  void intersection_assign(const Abs_Poly& y) override {
    ph.intersection_assign(get_poly(y));
  }
  void poly_hull_assign(const Abs_Poly& y) override {
    ph.poly_hull_assign(get_poly(y));
  }
  void poly_difference_assign(const Abs_Poly& y) override {
    ph.poly_difference_assign(get_poly(y));
  }
  void time_elapse_assign(const Abs_Poly& y) override {
    ph.time_elapse_assign(get_poly(y));
  }
  void topological_closure_assign() override {
    ph.topological_closure_assign();
  }
  void widening_assign(const Abs_Poly& y, widen_type widen_impl) override {
    ph.widening_assign(get_poly(y), widen_impl);
  }
  void widening_assign(const Abs_Poly& y, const Cons& cs,
                       widen_type widen_impl) override {
    ph.widening_assign(get_poly(y), cs, widen_impl);
  }
  Abs_Poly_Adapter* split(const Con& con, Topol topol) override {
    return new Abs_Poly_Adapter(ph.split(con, topol));
  }
  Abs_Poly_Adapter* split(const Con& c) override {
    return new Abs_Poly_Adapter(ph.split(c, topology()));
  }

  void
  affine_image(Var var, const Linear_Expr& expr,
               const Integer& inhomo = Integer::zero(),
               const Integer& den = Integer::one()) override {
    ph.affine_image(var, expr, inhomo, den);
  }
  void
  affine_preimage(Var var, const Linear_Expr& expr,
                  const Integer& inhomo = Integer::zero(),
                  const Integer& den = Integer::one()) override {
    ph.affine_preimage(var, expr, inhomo, den);
  }
  void
  parallel_affine_image(const Vars& vars,
                        const Linear_Exprs& exprs,
                        const Integers& inhomos,
                        const Integers& dens) override {
    ph.parallel_affine_image(vars, exprs, inhomos, dens);
  }

  void unconstrain(Var var) override { ph.unconstrain(var); }
  void unconstrain(const Index_Set& vars) override {
    ph.unconstrain(vars.begin(), vars.end());
  }
  void add_space_dims(dim_type d, bool project) override {
    ph.add_space_dims(d, project);
  }
  void map_space_dims(const Dims& pfunc) override {
    ph.map_space_dims(pfunc);
  }
  void remove_space_dim(Var var) override { ph.remove_space_dim(var); }
  void remove_space_dims(const Index_Set& vars) override {
    ph.remove_space_dims(vars.begin(), vars.end());
  }
  void remove_higher_space_dims(dim_type new_dim) override {
    ph.remove_higher_space_dims(new_dim);
  }
  void expand_space_dim(Var var, dim_type m) override {
    ph.expand_space_dim(var, m);
  }
  void fold_space_dims(const Index_Set& vars, Var dest) override{
    ph.fold_space_dims(vars, dest);
  }

  /* Input-output */
  void print(std::ostream& os) const override { ph.print(os); }
  void ascii_dump(std::ostream& s) const override { ph.ascii_dump(s); }

}; // Abs_Poly_Adapter

// Declarations of explicit instantiations of the available concrete classes.
extern template class Abs_Poly_Adapter<Poly>;
extern template class Abs_Poly_Adapter<Stats<Poly>>;
extern template class Abs_Poly_Adapter<F_Poly>;
extern template class Abs_Poly_Adapter<Stats<F_Poly>>;
extern template class Abs_Poly_Adapter<U_Poly>;
extern template class Abs_Poly_Adapter<Stats<U_Poly>>;

extern template class Abs_Poly_Adapter<U_Wrap<F_Poly>>;
extern template class Abs_Poly_Adapter<Stats<U_Wrap<F_Poly>>>;

// A poor man's object factory.
enum class Abs_Poly_Kind {
  POLY, POLY_STATS,
  F_POLY, F_POLY_STATS,
  U_POLY, U_POLY_STATS,
  UF_POLY, UF_POLY_STATS,
  SIZE
};

constexpr auto abs_poly_kind_size = static_cast<int>(Abs_Poly_Kind::SIZE);

constexpr const char* abs_poly_kind_names[abs_poly_kind_size] = {
  "Poly", "Poly_Stats",
  "F_Poly", "F_Poly_Stats",
  "U_Poly", "U_Poly_Stats",
  "UF_Poly", "UF_Poly_Stats"
};

extern PPLITE_TLS Abs_Poly_Kind abs_poly_kind;

inline const char* get_abs_poly_kind_name() {
  auto idx = static_cast<int>(abs_poly_kind);
  assert(0 <= idx && idx < abs_poly_kind_size);
  return abs_poly_kind_names[idx];
}

bool
abs_poly_name_to_kind(const char* cname,
                      Abs_Poly_Kind& kind,
                      bool& requires_stats);

// Returns true (and sets global variable accordingly)
// if "name" is the name of a (supported) concrete class.
bool set_abs_poly_kind(const char* name, bool noisy = false);

// Use this to actually build concrete objects
inline Abs_Poly*
make_poly(Abs_Poly_Kind kind, dim_type d, Spec_Elem s, Topol t) {
  switch (kind) {
  case Abs_Poly_Kind::POLY:
    return new Abs_Poly_Adapter<Poly>(d, s, t);
  case Abs_Poly_Kind::POLY_STATS:
    return new Abs_Poly_Adapter<Stats<Poly>>(d, s, t);
  case Abs_Poly_Kind::F_POLY:
    return new Abs_Poly_Adapter<F_Poly>(d, s, t);
  case Abs_Poly_Kind::F_POLY_STATS:
    return new Abs_Poly_Adapter<Stats<F_Poly>>(d, s, t);
  case Abs_Poly_Kind::U_POLY:
    return new Abs_Poly_Adapter<U_Poly>(d, s, t);
  case Abs_Poly_Kind::U_POLY_STATS:
    return new Abs_Poly_Adapter<Stats<U_Poly>>(d, s, t);
  case Abs_Poly_Kind::UF_POLY:
    return new Abs_Poly_Adapter<UF_Poly>(d, s, t);
  case Abs_Poly_Kind::UF_POLY_STATS:
    return new Abs_Poly_Adapter<Stats<UF_Poly>>(d, s, t);
  default:
    std::cerr << "Invalid Abs_Poly_Kind" << std::endl;
    assert(false);
    return nullptr;
  }
}

// Helper to use concrete kind names.
inline Abs_Poly*
make_poly(const char* kind_name, dim_type d, Spec_Elem s, Topol t) {
  auto kind = Abs_Poly_Kind::POLY;
  bool requires_stats = false;
  if (abs_poly_name_to_kind(kind_name, kind, requires_stats))
    return make_poly(kind, d, s, t);
  std::cerr << "Invalid Abs_Poly_Kind name" << std::endl;
  assert(false);
  return nullptr;
}

} // namespace dynamic
} // namespace pplite

#endif // pplite_Abs_Poly_Adapter_hh
