/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_BBox_hh
#define pplite_BBox_hh 1

#include "globals.hh"
#include "Rational.hh"
#include "Index_Set.hh"
#include "Gen.hh"

#include "Local_Stats.hh"

#include <vector>
#include <algorithm>
#include <iostream>
#include <cassert>

// FIXME: setting these to 0 requires rechecking old code.
#define PPLITE_BBOX_PSEUDO_VOLUME 1
#define PPLITE_BBOX_LESS_USES_ONLY_VOLUME 1

namespace pplite {

class BBox {
public: // Everything is public (meant!)
  struct Info {
    bool inf_lb = false;
    bool inf_ub = false;
    Rational lb;
    Rational ub;
    bool check_inv() const {
      return (inf_lb || inf_ub || lb <= ub)
        // Note: inf bound ==> finite bound == 0
        && (!inf_lb || lb.is_zero())
        && (!inf_ub || ub.is_zero());
    }

    bool is_bounded() const { return !inf_lb && !inf_ub; }
    Rational length() const {
      assert(is_bounded());
      return ub - lb;
    }
    int num_rays() const {
      return static_cast<int>(inf_lb) + static_cast<int>(inf_ub);
    }
    bool is_zero() const {
      return is_bounded() && lb.is_zero() && ub.is_zero();
    }
    void set_zero() {
      inf_lb = inf_ub = false;
      lb = Rational(0);
      ub = Rational(0);
    }

    bool contains(const Info& y) const {
      const auto& x = *this;
      assert(x.check_inv());
      assert(y.check_inv());
      return (x.inf_lb || (!y.inf_lb && x.lb <= y.lb))
        && (x.inf_ub || (!y.inf_ub && x.ub >= y.ub));
    }

    bool operator==(const Info& y) const {
      const auto& x = *this;
      assert(x.check_inv());
      assert(y.check_inv());
      return (x.inf_lb == y.inf_lb && x.inf_ub == y.inf_ub
              && (x.inf_lb || x.lb == y.lb)
              && (x.inf_ub || x.ub == y.ub));
    }

    // Note: x < y implies !x.contains(y)
    bool operator<(const Info& y) const {
      const auto& x = *this;
      if (y.inf_lb) {
        if (!x.inf_lb)
          return true;
        if (y.inf_ub)
          return !x.inf_ub;
        if (x.inf_ub)
          return false;
        return x.ub < y.ub;
      }
      if (x.inf_lb)
        return false;
      assert(!x.inf_lb && !y.inf_lb);
      switch (compare(x.lb, y.lb)) {
      case -1:
        return false;
      case 1:
        return true;
      case 0:
      default:
        if (x.inf_ub)
          return false;
        if (y.inf_ub)
          return true;
        assert(!x.inf_ub && !y.inf_ub);
        return x.ub < y.ub;
      }
    }

    // Returns true if refinement makes it empty.
    bool refine_as_integral() {
      if (!inf_lb) lb.round_up();
      if (!inf_ub) ub.round_down();
      if (is_bounded() && lb > ub) {
        set_zero();
        return true;
      }
      return false;
    }
  }; // struct Info

  // For volume in Volume_Info:
  // volume.first = number of infinite bounds (i.e., rays)
  // volume.second = *pseudo* volume
  using Volume_Info = std::pair<dim_type, Rational>;

  bool empty;
  Volume_Info volume;
  std::vector<Info> bounds;

public:
  BBox(const BBox& y) = default;
  BBox(BBox&& y) = default;
  BBox& operator=(const BBox& y) = default;
  BBox& operator=(BBox&& y) = default;
  ~BBox() = default;

  // The origin of the vector space: used for tag dispatching,
  // to avoid potential confusion with constructor of BBox universe.
  struct Origin {};
  static constexpr Origin ORIGIN = Origin();
  explicit BBox(dim_type sd, Origin)
    : empty(false),
      volume(0, Rational::zero()),
      bounds(sd) {}

  // Either the empty or the universe box.
  BBox(dim_type sd, Spec_Elem se = Spec_Elem::UNIVERSE)
    : BBox(sd, ORIGIN) {
    empty = (se == Spec_Elem::EMPTY);
    if (space_dim() == 0)
      volume.second = Rational::one();
    if (!empty) {
      volume.first = 2 * space_dim();
      for (auto& b : bounds) {
        b.inf_lb = true;
        b.inf_ub = true;
      }
    }
    assert(check_inv());
  }

  dim_type space_dim() const { return num_rows(bounds); }
  bool is_empty() const { return empty; }

  // Accessors for volume info.
  bool is_bounded() const { return volume.first == 0; }
  dim_type num_rays() const { return volume.first; }
  const Rational& pseudo_volume() const {
    assert(is_bounded());
    return volume.second;
  }

  // Accessors for bounds.
  bool& inf_lb(dim_type i) { return bounds[i].inf_lb; }
  bool& inf_ub(dim_type i) { return bounds[i].inf_ub; }
  Rational& lb(dim_type i) { return bounds[i].lb; }
  Rational& ub(dim_type i) { return bounds[i].ub; }
  bool inf_lb(dim_type i) const { return bounds[i].inf_lb; }
  bool inf_ub(dim_type i) const { return bounds[i].inf_ub; }
  const Rational& lb(dim_type i) const { return bounds[i].lb; }
  const Rational& ub(dim_type i) const { return bounds[i].ub; }
  Rational length(dim_type i) const { return bounds[i].length(); }

  bool contains(const BBox& y) const {
    auto& x = *this;
    assert(x.space_dim() == y.space_dim());
    if (y.is_empty())
      return true;
    if (x.is_empty())
      return false;
#if PPLITE_BBOX_PSEUDO_VOLUME
    if (x.is_bounded() && y.is_bounded()) {
      if (x.pseudo_volume() < y.pseudo_volume()) {
        // SAS2019 paper: Lemma 1.
        return false;
      }
    } else if (x.num_rays() < y.num_rays()) {
      // SAS2019 paper: Lemma 2.
      return false;
    }
#endif // PPLITE_BBOX_PSEUDO_VOLUME
    // Note: uses std::equal, but actually checks for containment.
    return std::equal(x.bounds.begin(), x.bounds.end(), y.bounds.begin(),
                      [](const BBox::Info& xi, const BBox::Info& yi) {
                        return xi.contains(yi);
                      });
  }

  bool equals(const BBox& y) const {
    const auto& x = *this;
    if (x.space_dim() != y.space_dim())
      return false;
    if (x.empty != y.is_empty())
      return false;
    if (x.empty)
      return true;
#if PPLITE_BBOX_PSEUDO_VOLUME
    if (x.is_bounded() && y.is_bounded()) {
      if (x.pseudo_volume() != y.pseudo_volume())
        return false;
    } else if (x.num_rays() != y.num_rays())
      return false;
#endif
    return (x.bounds == y.bounds);
  }

  bool is_included_in(const Con& c) const;

  bool is_disjoint_from(const BBox& y) const;

  // Modifiers:
  void set_empty() {
    empty = true;
#if PPLITE_BBOX_PSEUDO_VOLUME
    volume = { 0, Rational::zero() };
#endif
    std::for_each(bounds.begin(), bounds.end(),
                  std::mem_fn(&Info::set_zero));
    assert(check_inv());
  }

  void glb_assign(const BBox& y);
  void lub_assign(const BBox& y);
  void time_elapse_assign(const BBox& y);

  void add_gen(const Gen& g);
  template <typename Iter>
  void add_gens(Iter first, Iter last);
  void add_gens(const Gens& gs);
  void add_gens_as_rays(const Gens& gs);

  void unconstrain_lb(dim_type d) {
    assert(d < space_dim());
    if (inf_lb(d))
      return;
    inf_lb(d) = true;
    lb(d) = Rational::zero();
#if PPLITE_BBOX_PSEUDO_VOLUME
    ++volume.first;
    volume.second = Rational::zero();
#endif
  }
  void unconstrain_ub(dim_type d) {
    assert(d < space_dim());
    if (inf_ub(d))
      return;
    inf_ub(d) = true;
    ub(d) = Rational::zero();
#if PPLITE_BBOX_PSEUDO_VOLUME
    ++volume.first;
    volume.second = Rational::zero();
#endif
  }
  void unconstrain(dim_type d) {
    unconstrain_lb(d);
    unconstrain_ub(d);
  }
  void unconstrain(const Index_Set& vars) {
    unconstrain(vars.begin(), vars.end());
  }
  template <typename Iter>
  void unconstrain(Iter first, Iter last) {
    if (is_empty())
      return;
    while (first != last) {
      unconstrain(*first);
      ++first;
    }
  }

  Volume_Info
  compute_pseudo_volume() const {
    Volume_Info res = { 0, Rational::zero() };
#if PPLITE_BBOX_PSEUDO_VOLUME
    if (is_empty())
      return res;
    res.second = Rational::one();
    for (const auto& b : bounds) {
      auto nr = b.num_rays();
      if (nr == 0)
        res.second *= (Rational::one() + b.length());
      else {
        res.first += nr;
        res.second = Rational::zero();
      }
    }
#endif
    return res;
  }

  void recompute_pseudo_volume() {
    volume = compute_pseudo_volume();
  }

  // Note: x.less(y) implies !x.contains(y)
  bool less(const BBox& y) const {
#if PPLITE_BBOX_LESS_USES_ONLY_VOLUME
    const auto& x = *this;
    return x.volume < y.volume;
#else // !PPLITE_BBOX_LESS_USES_ONLY_VOLUME
    const auto& x = *this;
    if (x.bounded) {
      if (!y.bounded)
        return true;
      switch (compare(x.pseudo_volume, y.pseudo_volume)) {
      case -1:
        return true;
      case 1:
        return false;
      case 0:
      default:
        return x.bounds < y.bounds;
      }
    }
    assert(!x.bounded);
    if (y.bounded)
      return false;
    return x.bounds < y.bounds;
#endif // !PPLITE_BBOX_LESS_USES_ONLY_VOLUME
  }

  void affine_image(Var v, Linear_Expr ex, Integer inh, Integer d);

  void refine_as_integral(dim_type dim) {
    refine_as_integral(dim, dim+1);
  }
  void refine_as_integral(dim_type first, dim_type last);

  bool check_inv() const;
  void ascii_dump(std::ostream& os) const;

}; // class BBox

inline bool operator==(const BBox& x, const BBox& y) {
  return x.equals(y);
}
inline bool operator!=(const BBox& x, const BBox& y) {
  return !(x == y);
}

// Note: x < y implies !x.contains(y)
inline bool operator<(const BBox& x, const BBox& y) {
  return x.less(y);
}

} // namespace pplite

#endif // !defined(pplite_BBox_hh)
