/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Bit_Row_hh
#define pplite_Bit_Row_hh 1

#include "globals.hh"

#if PPLITE_LOW_LEVEL_COUNTERS
#include "Low_Level_Stats.hh"
#endif

#include <gmp.h>
#include <cassert>
#include <iosfwd>
#include <limits>

namespace pplite {

class Bit_Row;

void swap(Bit_Row& x, Bit_Row& y) noexcept;

bool operator==(const Bit_Row& x, const Bit_Row& y);
bool operator!=(const Bit_Row& x, const Bit_Row& y);
bool operator<(const Bit_Row& x, const Bit_Row& y);
int compare(const Bit_Row& x, const Bit_Row& y);
bool subset_or_equal(const Bit_Row& x, const Bit_Row& y);
bool subset_or_equal(const Bit_Row& x, const Bit_Row& y,
                     bool& strict_subset);
bool strict_subset(const Bit_Row& x, const Bit_Row& y);


class Bit_Row {
public:
  using Impl = mpz_t;

  Impl& impl() { return vec; }
  const Impl& impl() const { return vec; }

  Bit_Row() noexcept { mpz_init(vec); }
  Bit_Row(const Bit_Row& y) { mpz_init_set(vec, y.vec); }
  Bit_Row(Bit_Row&& y) noexcept : Bit_Row() { mpz_swap(vec, y.vec); }
  Bit_Row& operator=(const Bit_Row& y) {
    mpz_set(vec, y.vec);
    return *this;
  }
  Bit_Row& operator=(Bit_Row&& y) noexcept {
    mpz_swap(vec, y.vec);
    return *this;
  }
  ~Bit_Row() noexcept { mpz_clear(vec); }

  Bit_Row(const Bit_Row& y, const Bit_Row& z) {
    const auto y_size = y.vec->_mp_size;
    const auto z_size = z.vec->_mp_size;
    assert(y_size >= 0 && z_size >= 0);
    if (y_size < z_size) {
      assert(z_size <= max_size);
      mpz_init2(vec, z_size * bits_per_limb);
      union_helper(y, z);
    } else {
      assert(y_size <= max_size);
      mpz_init2(vec, y_size * bits_per_limb);
      union_helper(z, y);
    }
  }

  void m_swap(Bit_Row& y) noexcept { mpz_swap(vec, y.vec); }

  bool operator[](dim_type k) const {
    const auto vec_size = vec->_mp_size;
    assert(vec_size >= 0);
    const auto i = k / bits_per_limb;
    if (i >= vec_size)
      return false;
    const auto limb = *(vec->_mp_d + i);
    return ((limb >> (k % bits_per_limb)) & 1U) != 0;
  }

  void set(dim_type k) { mpz_setbit(vec, k); }
  void set_until(dim_type k);
  void clear(dim_type k) { mpz_clrbit(vec, k); }
  void clear_from(dim_type k) { mpz_tdiv_r_2exp(vec, vec, k); }
  void clear() { mpz_set_ui(vec, 0UL); }

  void complement_until(const dim_type k);

  void union_assign(const Bit_Row& x, const Bit_Row& y) {
    const auto x_size = x.vec->_mp_size;
    const auto y_size = y.vec->_mp_size;
    assert(x_size >= 0 && y_size >= 0);
    if (x_size < y_size) {
      assert(y_size <= max_size);
      mpz_realloc2(vec, y_size * bits_per_limb);
      union_helper(x, y);
    }
    else {
      assert(x_size <= max_size);
      mpz_realloc2(vec, x_size * bits_per_limb);
      union_helper(y, x);
    }
  }

  void intersection_assign(const Bit_Row& x, const Bit_Row& y) {
    mpz_and(vec, x.vec, y.vec);
  }

  void difference_assign(const Bit_Row& x, const Bit_Row& y) {
    mpz_t y_compl;
    mpz_init(y_compl);
    mpz_com(y_compl, y.vec);
    mpz_and(vec, x.vec, y_compl);
  }

  void shift_assign(dim_type shift) {
    if (shift > 0)
      mpz_mul_2exp(vec, vec, shift);
    else if (shift < 0)
      mpz_fdiv_q_2exp(vec, vec, -shift);
  }

  void copy_from(dim_type dst_pos,
                 const Bit_Row& src, dim_type src_pos, dim_type count);

  dim_type first() const;
  dim_type next(dim_type position) const;
  dim_type last() const;
  dim_type prev(dim_type position) const;

  bool empty() const { return mpz_sgn(vec) == 0; }
  bool intersects(const Bit_Row& y) const;

  dim_type count_ones() const {
#if PPLITE_LOW_LEVEL_COUNTERS
    LLOp_Stats::bump_sat_count();
#endif
    const auto x_size = vec->_mp_size;
    assert(x_size >= 0);
    return (x_size == 0) ? 0 : mpn_popcount(vec->_mp_d, x_size);
  }

  dim_type count_ones_in_union(const Bit_Row& y) const;

  bool check_inv() const;

  void ascii_dump(std::ostream& s) const;
  static void print_bits(mp_limb_t v);

  static constexpr dim_type bits_per_limb
  = sizeof_to_bits(PPLITE_SIZEOF_MP_LIMB_T);
  static_assert(bits_per_limb == static_cast<dim_type>(GMP_NUMB_BITS),
                "GMP using nail bits is not supported");
  static constexpr dim_type end_pos = std::numeric_limits<dim_type>::max();
  static constexpr dim_type max_size = end_pos / bits_per_limb;
  static constexpr mp_limb_t limb_all_set = ~static_cast<mp_limb_t>(0);

private:
  Impl vec;

  void union_helper(const Bit_Row& y, const Bit_Row& z);

  static dim_type
  first_one(const unsigned int u) { return __builtin_ctz(u); }
  static dim_type
  first_one(const unsigned long ul) { return __builtin_ctzl(ul); }
  static dim_type
  first_one(const unsigned long long ull) { return __builtin_ctzll(ull); }

  static dim_type last_one(const unsigned int u) {
    return sizeof_to_bits(sizeof(u)) - 1 - __builtin_clz(u);
  }
  static dim_type last_one(const unsigned long ul) {
    return sizeof_to_bits(sizeof(ul)) - 1 - __builtin_clzl(ul);
  }
  static dim_type last_one(const unsigned long long ull) {
    return sizeof_to_bits(sizeof(ull)) - 1 - __builtin_clzll(ull);
  }

  static mp_limb_t
  limb_shifted_copy(mp_limb_t src1, mp_limb_t src2, dim_type start) {
    assert(0 < start && start < bits_per_limb);
    mp_limb_t src_mask = (limb_all_set << start);
    return ((src1 & src_mask) >> start)
               | ((src2 & ~src_mask) << (limb_all_set - start + 1));
  }

  static mp_limb_t
  limb_partial_copy(mp_limb_t src, dim_type src_start,
                    mp_limb_t dst, dim_type dst_start,
                    dim_type count) {
    assert(count > 0);
    assert(src_start + count <= bits_per_limb);
    assert(dst_start + count <= bits_per_limb);
    mp_limb_t src_mask = (limb_all_set << src_start);
    if (src_start + count < bits_per_limb)
      src_mask &= ~(limb_all_set << (src_start + count));
    mp_limb_t dst_mask = ~(limb_all_set << dst_start);
    if (dst_start + count < bits_per_limb)
      dst_mask |= (limb_all_set << (dst_start + count));
    mp_limb_t src_masked_shifted = (src_start < dst_start)
      ? ((src & src_mask) << (dst_start - src_start))
      : ((src & src_mask) >> (src_start - dst_start));
    return (dst & dst_mask) | src_masked_shifted;
  }
}; // class Bit_Row

NOTHROW_DEFAULT_AND_MOVES(Bit_Row);

inline void
swap(Bit_Row& x, Bit_Row& y) noexcept { x.m_swap(y); }

inline bool
operator==(const Bit_Row& x, const Bit_Row& y) {
  const auto& x_vec = x.impl();
  const auto& y_vec = y.impl();
  const auto x_vec_size = x_vec->_mp_size;
  const auto y_vec_size = y_vec->_mp_size;
  assert(x_vec_size >= 0 && y_vec_size >= 0);
  if (x_vec_size != y_vec_size)
    return false;
  return mpn_cmp(x_vec->_mp_d, y_vec->_mp_d, x_vec_size) == 0;
}

inline bool operator!=(const Bit_Row& x, const Bit_Row& y) {
  return !(x == y);
}

inline bool
operator<(const Bit_Row& x, const Bit_Row& y) {
  return compare(x, y) < 0;
}

} // namespace pplite

#endif // !defined(pplite_Bit_Row_hh)
