/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_F_Poly_hh
#define pplite_F_Poly_hh 1

#include "globals.hh"
#include "Poly.hh"
#include "Poly_min.hh"
#include "Poly_templ.hh"

#include <cassert>
#include <iostream>
#include <iterator>
#include <numeric>
#include <vector>

namespace pplite {

class F_Poly {
public:
  using Block = std::vector<dim_type>;
  using Blocks = std::vector<Block>;
  using Factor = Poly;
  using Factors = std::vector<Factor>;

  // FIXME.
  Blocks& impl_blocks() { return blocks; }
  const Blocks& impl_blocks() const { return blocks; }
  Factors& impl_factors() { return factors; }
  const Factors& impl_factors() const { return factors; }

  static Topol get_default_topology() {
    return Poly::get_default_topology();
  }
  static void set_default_topology(Topol t) {
    return Poly::set_default_topology(t);
  }

  explicit
  F_Poly(dim_type d = 0, Spec_Elem s = Spec_Elem::UNIVERSE,
         Topol t = get_default_topology());
  F_Poly(dim_type d, Topol t, Spec_Elem s = Spec_Elem::UNIVERSE)
    : F_Poly(d, s, t) {}
  F_Poly(Spec_Elem s, dim_type d, Topol t = get_default_topology())
    : F_Poly(d, s, t) {}
  F_Poly(Spec_Elem s, Topol t, dim_type d)
    : F_Poly(d, s, t) {}
  F_Poly(Topol t, dim_type d, Spec_Elem s = Spec_Elem::UNIVERSE)
    : F_Poly(d, s, t) {}
  F_Poly(Topol t, Spec_Elem s, dim_type d)
    : F_Poly(d, s, t) {}

  F_Poly(const F_Poly& y) = default;
  F_Poly& operator=(const F_Poly& y) = default;
  F_Poly(F_Poly&& y) = default;
  F_Poly& operator=(F_Poly&& y) = default;
  ~F_Poly() = default;

  // Helper for debugging.
  explicit F_Poly(const Poly& ph);

  /* Predicates */
  bool check_inv() const;
  bool is_necessarily_closed() const { return topology() == Topol::CLOSED; }
  bool is_empty() const { return empty; }
  bool is_minimized() const {
    return is_empty()
      || std::all_of(factors.begin(), factors.end(),
                     std::mem_fn(&Poly::is_minimized));
  }
  bool is_universe() const {
    return !is_empty()
      && std::all_of(factors.begin(), factors.end(),
                     std::mem_fn(&Poly::is_universe));
  }
  bool is_topologically_closed() const {
    return is_necessarily_closed()
      || is_empty()
      || std::all_of(factors.begin(), factors.end(),
                     std::mem_fn(&Poly::is_topologically_closed));
  };
  bool is_bounded() const {
    return is_empty()
      || std::all_of(factors.begin(), factors.end(),
                     std::mem_fn(&Poly::is_bounded));
  }
  bool is_bounded_expr(bool from_below, const Linear_Expr& expr) const;
  bool constrains(Var var) const;
  bool equals(const F_Poly& y) const;
  bool contains(const F_Poly& y) const;
  bool strictly_contains(const F_Poly& y) const;
  bool is_disjoint_from(const F_Poly& y) const;
  bool boxed_contains(const F_Poly& y) const {
    // FIXME: probably not worth more effort.
    return contains(y);
  }

  /* Queries */
  Topol topology() const { return topol; }
  dim_type space_dim() const { return dim; }
  dim_type affine_dim() const;
  Poly_Con_Rel relation_with(const Con& c) const;
  Poly_Gen_Rel relation_with(const Gen& g) const;
  bool min_max(const bool minimizing,
               const Linear_Expr& expr, const Integer& inhomo,
               Integer& num, Integer& den, bool& included, Gen& g) const;
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(true, expr, inhomo, num, den, included, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(false, expr, inhomo, num, den, included, g);
  }
  bool get_bounds(Var var,
                  Rational& min, Rational& max,
                  bool& min_inf, bool& max_inf) const;

  size_t hash() const;

  void print(std::ostream& os) const;
  void print() const { print(std::cout); }

  using Cons_Proxy = Cont_Proxy<Cons>;
  Cons_Proxy cons() const { return Cons_Proxy(copy_cons()); }
  using Gens_Proxy = Cont_Proxy<Gens>;
  Gens_Proxy gens() const { return Gens_Proxy(copy_gens()); }
  Cons copy_cons() const;
  Gens copy_gens() const;
  Cons_Proxy normalized_cons() const { return Cons_Proxy(copy_normalized_cons()); }
  Cons copy_normalized_cons() const;
  using Gens_Info = Poly::Gens_Info;
  Gens_Info gens_info() const;
  dim_type num_min_cons() const;
  dim_type num_min_gens() const;

  BBox get_bounding_box() const;

  /* Modifiers */
  void m_swap(F_Poly& y) noexcept;
  void set_empty();
  void set_universe();
  void set_topology(Topol t) {
    if (topology() != t) {
      topol = t;
      for (auto& f : factors)
        f.set_topology(t);
    }
  }

  void add_con(const Con& c);
  void add_cons(const Cons& cs);
  template <typename Iter>
  void add_cons(Iter first, Iter last, bool do_normalize = true);

  void add_gen(const Gen& g) {
    add_gens(&g, &g + 1);
  }
  void add_gens(const Gens& gs) {
    add_gens(gs.begin(), gs.end());
  }
  template <typename Iter>
  void add_gens(Iter first, Iter last) {
    if (first == last)
      return;
    auto p_iter = std::find_if(first, last, std::mem_fn(&Gen::is_point));
    if (p_iter != last) {
      Poly ph(space_dim(), Spec_Elem::EMPTY, topology());
      ph.add_gens(first, last);
      F_Poly y(ph);
      y.factorize();
      poly_hull_assign(y);
      return;
    }
    // No point in the generators.
    assert(!is_empty());
    auto c_iter = std::find_if(first, last,
                               std::mem_fn(&Gen::is_closure_point));
    if (c_iter != last) {
      Factor ph = get_poly();
      ph.add_gens(first, last);
      *this = F_Poly(ph);
      factorize();
      return;
    }
    // All gens are lines or rays.
    for ( ; first != last; ++first)
      add_line_or_ray(*first);
    factorize();
  }

  void topological_closure_assign() {
    if (is_necessarily_closed())
      return;
    if (is_empty())
      return;
    for (auto& f : factors)
      f.topological_closure_assign();
  }

  template <typename Iter>
  void unconstrain(Iter first, Iter last) {
    for ( ; first != last; ++first)
      unconstrain(Var(*first));
  }
  void unconstrain(const Index_Set& vars) {
    unconstrain(vars.begin(), vars.end());
  }
  void unconstrain(Var var);

  void intersection_assign(const F_Poly& y);
  void poly_hull_assign(const F_Poly& y);
  void con_hull_assign(const F_Poly& y, bool boxed = false) {
    con_hull(*this, &y, &y + 1, boxed);
  }
  void poly_difference_assign(const F_Poly&);

  void affine_image(Var var, const Linear_Expr& expr,
                    const Integer& inhomo = Integer::zero(),
                    const Integer& den = Integer::one());
  void affine_preimage(Var var, const Linear_Expr& expr,
                       const Integer& inhomo = Integer::zero(),
                       const Integer& den = Integer::one());
  void
  parallel_affine_image(const Vars& vars,
                        const Linear_Exprs& exprs,
                        const Integers& inhomos, const Integers& dens);

  using widen_type = typename Poly::widen_type;
  static widen_type get_widen_impl() {
    return Poly::get_widen_impl();
  }
  static void set_widen_impl(widen_type widen) {
    Poly::set_widen_impl(widen);
  }

  void widening_assign(const F_Poly& y, widen_type widen,
                       const Cons* upto_ptr);
  void widening_assign(const F_Poly& y,
                       widen_type widen = get_widen_impl()) {
    widening_assign(y, widen, nullptr);
  }
  void widening_assign(const F_Poly& y, const Cons& upto_cons,
                       widen_type widen = get_widen_impl()) {
    widening_assign(y, widen, &upto_cons);
  }

  void time_elapse_assign(const F_Poly&);
  void positive_time_elapse_assign(const F_Poly&) {
    // FIXME
    assert(false);
  }

  F_Poly split(const Con& c, Topol t);
  F_Poly split(const Con& c) { return split(c, topology()); }

  /* Change of space dim */
  void add_space_dims(dim_type m, bool project = false);
  void map_space_dims(const Dims& pfunc);
  void remove_higher_space_dims(dim_type new_dim);

  template <typename Iter>
  void remove_space_dims(Iter first, Iter last) {
    if (is_empty()) {
      dim -= std::distance(first, last);
      return;
    }
    // FIXME: partition the vars to be removed according to blocks
    // and implement removal for each block.
    for (dim_type i = space_dim(); i-- >= 0; ) {
      auto tbr_dim = std::find(first, last, i);
      if (tbr_dim != last)
        remove_space_dim(Var(i));
    }
  }
  void remove_space_dims(const Index_Set& vars) {
    remove_space_dims(vars.begin(), vars.end());
  }
  void remove_space_dim(Var var);

  void concatenate_assign(const F_Poly& y);
  void expand_space_dim(Var var, dim_type m);
  void fold_space_dims(const Index_Set& vars, Var dest);

  // Semantically const, but may affect syntactic representation.
  void minimize() const;

  /* Input-output */
  bool ascii_load(std::istream& s);
  void ascii_dump(std::ostream& s) const;

  Poly get_poly() const;

private:
  dim_type dim;
  bool empty;
  Topol topol;
  Blocks blocks;
  Factors factors;

  bool is_normalized;

  dim_type merge(const Block& b);

  // Semantically const, but may affect syntactic representation.
  void sync(const Blocks& bs) const;
  void factorize() const;

  void normalize() const;

  template <typename Hull_Fun>
  void hull_assign(const F_Poly& y, Hull_Fun hull_fun);

  void add_line_or_ray(const Gen& g);
  void remove_space_dim(const dim_type v);

}; // class F_Poly

namespace IO_Operators {

inline std::ostream&
operator<<(std::ostream& os, const F_Poly& ph) {
  ph.print(os);
  return os;
}

} // namespace IO_Operators

inline void
swap(F_Poly& x, F_Poly& y) noexcept { x.m_swap(y); }

inline bool
operator==(const F_Poly& x, const F_Poly& y) { return x.equals(y); }
inline bool
operator!=(const F_Poly& x, const F_Poly& y) { return !(x == y); }

} // namespace pplite

#endif // !defined(pplite_F_Poly_hh)
