/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Index_Partition_hh
#define pplite_Index_Partition_hh 1

namespace pplite {
namespace detail {

enum class Range { EQ, POS, NEG, POS_NEG };

using Ranges = std::vector<Range>;

template <Range policy>
class Index_Partition {
  const Ranges& labels;
public:
  Index_Partition() = delete;
  Index_Partition(const Index_Partition&) = default;
  Index_Partition& operator=(const Index_Partition&) = default;
  ~Index_Partition() = default;

  // Not movable
  Index_Partition(Index_Partition&&) = delete;
  Index_Partition& operator=(Index_Partition&&) = delete;

  explicit Index_Partition(const Ranges& ls) : labels(ls) {}

  dim_type tot_size() const { return labels.size(); }
  Range operator[](dim_type i) const { return labels[i]; }

  // read-only forward iterator for `policy' labels
  template <Range policy_it>
  class iterator {
  public:
    using iterator_category = std::forward_iterator_tag;
    using value_type = dim_type;
    using pointer = const value_type*;
    using reference = const value_type&;
    using difference_type = std::ptrdiff_t;

  private:
    const Index_Partition& part;
    dim_type pos;
    dim_type initial_end_pos;

    dim_type end_pos() const {
      // New EQ generators may have been added
      // check the actual size of part.
      // FIXME: Enea: it is unclear what is really meant here.
      // FIXME: the partition should be logically constant during iteration.
      // FIXME: why this special case for EQ?
      if (policy == Range::EQ)
        return part.tot_size();
      // If policy is POS or NEG, keep the initial end position.
      return initial_end_pos;
    }

  public:
    iterator(const Index_Partition& partition,
             bool get_first)
      : part(partition), initial_end_pos(part.tot_size()) {
      if (!get_first){
        pos = end_pos();
        return;
      }
      // get_first
      for (pos = 0; pos < end_pos(); ++pos){
        if (part[pos] == policy_it)
          break;
      }
    }

    value_type operator*() const { return pos; }
    iterator operator++() {
      if (pos == end_pos()) return *this;
      do { ++pos; }
      while (pos != end_pos() && part[pos] != policy_it);
      return *this;
    }
    bool operator==(const iterator& i) const { return pos == i.pos; }
    bool operator!=(const iterator& i) const { return pos != i.pos; }
  }; // iterator

  iterator<policy> begin() const {
    return iterator<policy>(*this, true);
  }
  iterator<policy> end() const {
    return iterator<policy>(*this, false);
  }
  bool empty() const { return begin() == end(); }
}; // class Index_Partition

template <Range policy>
class Sparse_Index_Partition {
public:
  using Index = dim_type;
  using Indices = std::vector<Index>;

  Sparse_Index_Partition() = default;
  Sparse_Index_Partition(const Sparse_Index_Partition&) = delete;
  Sparse_Index_Partition(Sparse_Index_Partition&&) = delete;
  Sparse_Index_Partition& operator=(const Sparse_Index_Partition&) = delete;
  Sparse_Index_Partition& operator=(Sparse_Index_Partition&&) = delete;
  ~Sparse_Index_Partition() = default;

  explicit Sparse_Index_Partition(const Ranges& r) { reset(r); }

  void reset(const Ranges& r) {
    vec.reserve(std::count(r.begin(), r.end(), policy));
    vec.clear();
    for (Index i = 0, i_end = r.size(); i != i_end; ++i) {
      if (r[i] == policy)
        vec.push_back(i);
    }
  }

  Indices::const_iterator begin() const { return vec.begin(); }
  Indices::const_iterator end() const { return vec.end(); }
  dim_type size() const { return vec.size(); }
private:
  Indices vec;
}; // class Sparse_Index_Partition

} // namespace detail
} // namespace pplite

#endif  // not defined pplite_Index_Partition_hh
