/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Index_Set_hh
#define pplite_Index_Set_hh 1

#include "globals.hh"
#include "Bit_Row.hh"
#include <algorithm>
#include <cassert>
#include <iostream>
#include <iterator>
#include <utility>
#include <vector>

namespace pplite {

class Index_Set;

bool operator==(const Index_Set& x, const Index_Set& y);
bool operator!=(const Index_Set& x, const Index_Set& y);
bool subset_or_equal(const Index_Set& x, const Index_Set& y);
bool strict_subset(const Index_Set& x, const Index_Set& y);

class Index_Set {
  Bit_Row bits;
public:
  using Impl = Bit_Row;
  Impl& impl() { return bits; }
  const Impl& impl() const { return bits; }

  Index_Set() = default;
  Index_Set(const Index_Set&) = default;
  Index_Set(Index_Set&&) = default;
  Index_Set& operator=(const Index_Set&) = default;
  Index_Set& operator=(Index_Set&&) = default;
  ~Index_Set() = default;

  explicit Index_Set(const dim_type i)
    : Index_Set() { set(i); }
  explicit Index_Set(const dim_type i, const dim_type j)
    : Index_Set(std::max(i, j)) { set(std::min(i, j)); }

  explicit Index_Set(Bit_Row&& row) : bits(std::move(row)) { }

  //! Set-union constructor (needed?).
  Index_Set(const Index_Set& y, const Index_Set& z);

  // Provide a read-only forward iterator on the indices in the set.
  class const_iterator {
  public:
    using iterator_category = std::bidirectional_iterator_tag;
    using value_type = dim_type;
    using pointer = const value_type*;
    using reference = const value_type&;
    using difference_type = std::ptrdiff_t;

  private:
    const Index_Set* iset_ptr;
    value_type pos;
    static constexpr value_type end_pos = Index_Set::Impl::end_pos;

  public:
    const_iterator() = default;
    const_iterator(const const_iterator&) = default;
    const_iterator(const_iterator&&) = default;
    const_iterator& operator=(const const_iterator&) = default;
    const_iterator& operator=(const_iterator&&) = default;
    ~const_iterator() = default;

    const_iterator(const Index_Set* is, bool get_first)
      : iset_ptr(is),
        pos(get_first ? is->impl().first() : end_pos) { }

    value_type operator*() const {
      assert(pos != end_pos);
      return pos;
    }

    const_iterator& operator++() {
      assert(pos != end_pos);
      pos = iset_ptr->impl().next(pos);
      return *this;
    }

    const_iterator& operator--() {
      assert(pos != iset_ptr->impl().first());
      pos = (pos == end_pos)
        ? iset_ptr->impl().last()
        : iset_ptr->impl().prev(pos);
      return *this;
    }

    bool operator==(const const_iterator& j) const {
      assert(iset_ptr == j.iset_ptr);
      return pos == j.pos;
    }

    bool operator!=(const const_iterator& j) const {
      return !(*this == j);
    }

  }; // class const_iterator

  NOTHROW_DEFAULT_AND_MOVES(const_iterator);

  const_iterator begin() const { return const_iterator(this, true); }
  const_iterator end() const { return const_iterator(this, false); }

  bool test(dim_type y) const { return bits[y]; }
  void set(dim_type y) { bits.set(y); }
  void set_until(dim_type y) { bits.set_until(y); }
  void clear(dim_type y) { bits.clear(y); }
  void clear() { bits.clear(); }

  bool empty() const { return bits.empty(); }
  dim_type size() const { return bits.count_ones(); }
  dim_type first() const { return bits.first(); }
  dim_type last() const { return bits.last(); }
  bool intersects(const Index_Set& y) const {
    return bits.intersects(y.bits);
  }

  void union_assign(const Index_Set& y) {
    bits.union_assign(bits, y.bits);
  }
  void intersection_assign(const Index_Set& y) {
    bits.intersection_assign(bits, y.bits);
  }
  void difference_assign(const Index_Set& y) {
    bits.difference_assign(bits, y.bits);
  }
  void shift_assign(dim_type shift) {
    bits.shift_assign(shift);
  }

  template <typename Iter>
  void remove_all(Iter rem_first, Iter rem_last);
  template <typename Indices>
  void remove_all(const Indices& indices) {
    remove_all(std::begin(indices), std::end(indices));
  }
  // For debugging purposes.
  template <typename Iter>
  Index_Set remove_all_aux(Iter rem_first, Iter rem_last) const;

  void ascii_dump(std::ostream& os) const;
  bool ascii_load(std::istream& is);

  void print(std::ostream& os) const { ascii_dump(os); }
  void print() const { print(std::cout); }

}; // class Index_Set

NOTHROW_DEFAULT_AND_MOVES(Index_Set);

using NS_Rows = std::vector<Index_Set>;

inline bool
operator==(const Index_Set& x, const Index_Set& y) {
  return x.impl() == y.impl();
}

inline bool
operator!=(const Index_Set& x, const Index_Set& y) {
  return x.impl() != y.impl();
}

inline bool
operator<(const Index_Set& x, const Index_Set& y) {
  return x.impl() < y.impl();
}

inline bool
subset_or_equal(const Index_Set& x, const Index_Set& y) {
  return subset_or_equal(x.impl(), y.impl());
}

inline bool
strict_subset(const Index_Set& x, const Index_Set& y) {
  return strict_subset(x.impl(), y.impl());
}

inline void
swap(Index_Set& x, Index_Set& y) noexcept {
  x.impl().m_swap(y.impl());
}


template <typename Iter>
Index_Set
Index_Set::remove_all_aux(Iter rem_first, Iter rem_last) const {
  const Index_Set& iset = *this;
  Index_Set res;
  if (iset.empty())
    return res;
  if (rem_first == rem_last) {
    res = iset;
    return res;
  }
  auto i = iset.begin(), i_end = iset.end();
  auto j = rem_first, j_end = rem_last;
  dim_type removed = 0;
  while (i != i_end && j != j_end) {
    auto idx_i = *i;
    auto idx_j = *j;
    if (idx_i < idx_j) {
      res.set(idx_i - removed);
      ++i;
      continue;
    } else if (idx_i == idx_j) {
      ++i;
      ++j;
      ++removed;
    } else {
      ++j;
      ++removed;
    }
  }
  for ( ; i != i_end; ++i) {
    res.set(*i - removed);
  }
  return res;
}

template <typename Iter>
void
Index_Set::remove_all(Iter rem_first, Iter rem_last) {
  Bit_Row& row = bits;
  if (row.empty())
    return;
  if (rem_first == rem_last)
    return;
  dim_type row_last = row.last();
  assert(row_last != Impl::end_pos);
  // No need to copy first chunk.
  dim_type dst_pos = *rem_first;
  dim_type src_pos = dst_pos + 1;
  if (src_pos > row_last) {
    row.clear_from(dst_pos);
    return;
  }
  ++rem_first;
  for ( ; rem_first != rem_last; ++rem_first) {
    auto rem_pos = std::min(*rem_first, row_last + 1);
    assert(rem_pos >= src_pos);
    auto count = rem_pos - src_pos;
    if (count > 0) {
      row.copy_from(dst_pos, row, src_pos, count);
      dst_pos += count;
      src_pos += count;
    }
    ++src_pos;
    if (src_pos > row_last) {
      row.clear_from(dst_pos);
      return;
    }
  }
  // Copy suffix (if any).
  assert(row_last >= src_pos);
  auto count = (row_last + 1) - src_pos;
  if (count > 0) {
    row.copy_from(dst_pos, row, src_pos, count);
  }
  // Clear trailing bits.
  row.clear_from(dst_pos + count);
}

} // namespace pplite

#endif // !defined(pplite_Index_Set_hh)
