/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_clock_hh
#define pplite_clock_hh 1

#include "globals.hh"

#include <ctime>
#include <iostream>
#include <iomanip>
#include <sys/resource.h>
#include <utility>

namespace pplite {

// This is a simple clock: it starts automatically when created;
// it can be restarted using restart(); it can be queried by calling
// elapsed_time(), returning the time elapsed since last restart,
// in seconds and microseconds (a pair).
class Clock {
public:
  Clock() { restart(); }
  // Not copyable, only moveable.
  Clock(const Clock&) = delete;
  Clock& operator=(const Clock&) = delete;
  Clock(Clock&&) = default;
  Clock& operator=(Clock&&) = default;
  ~Clock() = default;

  void restart();
  std::pair<time_t, time_t> elapsed_time() const;
  void print_elapsed(std::ostream& os) const { print(os, elapsed_time()); }

  static void print(std::ostream& os, time_t secs, time_t usecs) {
    const char fill_char = os.fill();
    os << secs << "." << std::setfill('0') << std::setw(6) << usecs;
    os.fill(fill_char);
  }
  static void print(std::ostream& os, const std::pair<time_t, time_t>& p) {
    print(os, p.first, p.second);
  }

private:
#define USE_RUSAGE 1
#if USE_RUSAGE
  using Saved_Time = struct timeval;
#else
  using Saved_Time = struct timespec;
#endif
  Saved_Time saved_time;
}; // class Clock

// A data structure collecting simple time stats.
struct Time_Stats {
  unsigned long counter; // Number of calls
  time_t secs;           // Cumulative seconds spent in calls
  time_t usecs;          // Cumulative microseconds spent in calls
  Time_Stats() : counter(0), secs(0), usecs(0) {}

  // Adds `incr_secs' seconds and `incr_usecs' microseconds to stats,
  // increasing by 1 the number of calls.
  void incr(time_t incr_secs, time_t incr_usecs) {
    ++counter;
    secs += incr_secs;
    usecs += incr_usecs;
    if (usecs >= 1000000) {
      ++secs;
      usecs %= 1000000;
    }
  }
}; // struct Time_Stats

} // namespace pplite

#endif // !defined(pplite_clock_hh)
