/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_utils_hh
#define pplite_utils_hh 1

#include "globals.hh"

#include <iterator>
#include <functional>
#include <vector>
#include <list>
#include <cassert>

namespace pplite {

// This helper function returns a dim_type value:
// used to avoid warnings about signed/unsigned comparisons.
template <typename Container>
inline dim_type
num_rows(const Container& rows) { return rows.size(); }

template <typename Row, typename Iter>
inline void
erase_using_sorted_indices(std::vector<Row>& rows,
                           Iter first, Iter last) {
  if (first == last)
    return;

  // `idx_unused' and `idx' start with value `*first' (instead of `0'),
  // because the loop would just increment them in the preceding iterations.
  using Index = typename std::iterator_traits<Iter>::value_type;
  const Index idx_end = rows.size();
  auto idx_unused = *first;
  assert(0 <= idx_unused);
  if (idx_unused >= idx_end)
    return;
  auto idx = idx_unused + 1;
  ++first;
  while (first != last && idx != idx_end) {
    auto new_idx = *first;
    assert(idx_unused < idx && idx <= new_idx);
    if (idx == new_idx) {
      // Don't swap row idx.
      ++first;
      ++idx;
    } else {
      // Keep (i.e., swap to its correct place) row idx.
      using std::swap;
      swap(rows[idx_unused], rows[idx]);
      ++idx_unused;
      ++idx;
    }
  }
  // Move up the remaining rows, if any.
  for ( ; idx < idx_end; ++idx) {
    using std::swap;
    swap(rows[idx_unused], rows[idx]);
    ++idx_unused;
  }
  // The trailing rows have to be erased.
  rows.resize(idx_unused);
}

template <typename Row, typename Indices>
inline void
erase_using_sorted_indices(std::vector<Row>& rows, const Indices& indices) {
  erase_using_sorted_indices(rows, std::begin(indices), std::end(indices));
}

inline size_t
hash_size(size_t val) {
  std::hash<size_t> hash_fn;
  return hash_fn(val);
}

/* Boost code (GPL compatible) */
inline void
hash_combine(size_t& lhs, size_t rhs) {
  lhs ^= rhs + 0x9e3779b9 + (lhs << 6) + (lhs >> 2);
}

/* Adapter for a sequence of elements into a sequence of Val_Type. */
template <typename Dst_Type>
struct SequenceAdapter {
  using value_type = Dst_Type;
  using List = typename std::list<const value_type*>;
  using list_const_it = typename List::const_iterator;

  struct const_iterator {
    using value_type = Dst_Type;
    using pointer = const value_type*;
    using reference = const value_type&;
    using difference_type = typename list_const_it::difference_type;
    using iterator_category = typename list_const_it::iterator_category;

    list_const_it it;

    const_iterator(list_const_it i) : it(i) {}
    reference operator*() const         { return *(*it); }
    pointer operator->() const          { return (*it); }
    bool operator==(const_iterator y)   { return it == y.it; }
    bool operator!=(const_iterator y)   { return it != y.it; }
    const_iterator operator++()         { ++it; return *this; }
  };

  struct TrueFunc {
    template <typename Src_Type>
    bool operator()(const Src_Type&) { return true; }
  };
  struct IdAddr {
    const Dst_Type* operator()(const Dst_Type& x) { return &x; }
  };

  template <typename Iter,
            typename Src_Type = typename std::iterator_traits<Iter>::value_type,
            typename CondFunc = std::function< bool (const Src_Type&)>,
            typename Reinterpreter = std::function< const Dst_Type* (const Src_Type&)>>
  SequenceAdapter(Iter first, Iter last, CondFunc cond,
                  Reinterpreter get_ptr = IdAddr()) {
    for (auto i = first; i != last; ++i) {
      if (cond(*i))
        ptrs.push_back(get_ptr(*i));
    }
  }

  bool empty() const                    { return ptrs.empty(); }
  dim_type size() const                 { return ptrs.size(); }
  const_iterator begin() const          { return ptrs.begin(); }
  const_iterator end() const            { return ptrs.end(); }
  void push_back(const value_type* p)   { ptrs.push_back(p); }

private:
  List ptrs;
};

} // namespace pplite

#endif // !defined(pplite_utils_hh)
